"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const results_1 = require("@expo/results");
const ReadonlyEntity_1 = __importDefault(require("./ReadonlyEntity"));
/**
 * Entity is a privacy-first data model.
 *
 * A instance of an entity represents a single "row" of persisted data in a database that a
 * viewer, represented by the corresponding {@link ViewerContext}, has permission to read.
 *
 * Create, read, update, and delete permissions for an entity are declaratively defined using an
 * {@link EntityPrivacyPolicy}.
 *
 * Entites are loaded through an {@link EntityLoader}, which is responsible for
 * orchestrating fetching, caching, and authorization of reading "rows".
 *
 * Entities are mutated and deleted through an {@link EntityMutator}, which is responsible for
 * orchestrating database writes, cache invalidation, and authorization of writing "rows".
 *
 * All concrete entity implementations should extend this class and provide their
 * own {@link EntityCompanionDefinition}.
 */
class Entity extends ReadonlyEntity_1.default {
    /**
     * Vend mutator for creating a new entity in given query context.
     * @param viewerContext - viewer context of creating user
     * @param queryContext - query context in which to perform the create
     * @returns mutator for creating an entity
     */
    static creator(viewerContext, queryContext = viewerContext
        .getViewerScopedEntityCompanionForClass(this)
        .getQueryContextProvider()
        .getQueryContext()) {
        return viewerContext
            .getViewerScopedEntityCompanionForClass(this)
            .getMutatorFactory()
            .forCreate(queryContext);
    }
    /**
     * Vend mutator for updating an existing entity in given query context.
     * @param existingEntity - entity to update
     * @param queryContext - query context in which to perform the update
     * @returns mutator for updating existingEntity
     */
    static updater(existingEntity, queryContext = existingEntity
        .getViewerContext()
        .getViewerScopedEntityCompanionForClass(this)
        .getQueryContextProvider()
        .getQueryContext()) {
        return existingEntity
            .getViewerContext()
            .getViewerScopedEntityCompanionForClass(this)
            .getMutatorFactory()
            .forUpdate(existingEntity, queryContext);
    }
    /**
     * Delete an existing entity in given query context.
     * @param existingEntity - entity to delete
     * @param queryContext - query context in which to perform the delete
     */
    static deleteAsync(existingEntity, queryContext = existingEntity
        .getViewerContext()
        .getViewerScopedEntityCompanionForClass(this)
        .getQueryContextProvider()
        .getQueryContext()) {
        return existingEntity
            .getViewerContext()
            .getViewerScopedEntityCompanionForClass(this)
            .getMutatorFactory()
            .forDelete(existingEntity, queryContext)
            .deleteAsync();
    }
    /**
     * Delete an existing entity in given query context, throwing if deletion is unsuccessful.
     * @param existingEntity - entity to delete
     * @param queryContext - query context in which to perform the delete
     */
    static enforceDeleteAsync(existingEntity, queryContext = existingEntity
        .getViewerContext()
        .getViewerScopedEntityCompanionForClass(this)
        .getQueryContextProvider()
        .getQueryContext()) {
        return existingEntity
            .getViewerContext()
            .getViewerScopedEntityCompanionForClass(this)
            .getMutatorFactory()
            .forDelete(existingEntity, queryContext)
            .enforceDeleteAsync();
    }
    /**
     * Check whether an entity loaded by a viewer can be updated by that same viewer.
     *
     * @remarks
     *
     * This may be useful in situations relying upon the thrown privacy policy thrown authorization error
     * is insufficient for the task at hand. When dealing with purely a sequence of mutations it is easy
     * to roll back all mutations given a single authorization error by wrapping them in a single transaction.
     * When certain portions of a mutation cannot be rolled back transactionally (third pary calls,
     * legacy code, etc), using this method can help decide whether the sequence of mutations will fail before
     * attempting them. Note that if any privacy policy rules use a piece of data being updated in the mutations
     * the result of this method and the update mutation itself may differ.
     *
     * @param existingEntity - entity loaded by viewer
     * @param queryContext - query context in which to perform the check
     */
    static async canViewerUpdateAsync(existingEntity, queryContext = existingEntity
        .getViewerContext()
        .getViewerScopedEntityCompanionForClass(this)
        .getQueryContextProvider()
        .getQueryContext()) {
        const privacyPolicy = new (this.getCompanionDefinition().privacyPolicyClass)();
        const evaluationResult = await results_1.asyncResult(privacyPolicy.authorizeUpdateAsync(existingEntity.getViewerContext(), queryContext, existingEntity));
        return evaluationResult.ok;
    }
    /**
     * Check whether an entity loaded by a viewer can be deleted by that same viewer.
     *
     * @remarks
     * See remarks for canViewerUpdate.
     *
     * @param existingEntity - entity loaded by viewer
     * @param queryContext - query context in which to perform the check
     */
    static async canViewerDeleteAsync(existingEntity, queryContext = existingEntity
        .getViewerContext()
        .getViewerScopedEntityCompanionForClass(this)
        .getQueryContextProvider()
        .getQueryContext()) {
        const privacyPolicy = new (this.getCompanionDefinition().privacyPolicyClass)();
        const evaluationResult = await results_1.asyncResult(privacyPolicy.authorizeDeleteAsync(existingEntity.getViewerContext(), queryContext, existingEntity));
        return evaluationResult.ok;
    }
}
exports.default = Entity;
//# sourceMappingURL=Entity.js.map