import { IEntityClass } from './Entity';
import EntityCompanion, { IPrivacyPolicyClass } from './EntityCompanion';
import EntityConfiguration from './EntityConfiguration';
import EntityMutationTriggerConfiguration from './EntityMutationTriggerConfiguration';
import EntityMutationValidator from './EntityMutationValidator';
import EntityPrivacyPolicy from './EntityPrivacyPolicy';
import EntityQueryContextProvider from './EntityQueryContextProvider';
import IEntityCacheAdapterProvider from './IEntityCacheAdapterProvider';
import IEntityDatabaseAdapterProvider from './IEntityDatabaseAdapterProvider';
import ReadonlyEntity from './ReadonlyEntity';
import ViewerContext from './ViewerContext';
import IEntityMetricsAdapter from './metrics/IEntityMetricsAdapter';
/**
 * Backing database and transaction type for an entity. The definitions and implementations
 * are provided by injection in the root EntityCompanionProvider to allow for mocking and sharing.
 */
export declare type DatabaseAdapterFlavor = string;
/**
 * Cache system for an entity. The definitions and implementations are provided by injection
 * in the root EntityCompanionProvider to allow for mocking and sharing.
 */
export declare type CacheAdapterFlavor = string;
/**
 * Defines a set interfaces for a entity database adapter flavor. An entity that uses a flavor
 * will use the specified adapter for database accesses and the specified query context provider
 * for providing query contexts.
 */
export interface DatabaseAdapterFlavorDefinition {
    adapterProvider: IEntityDatabaseAdapterProvider;
    queryContextProvider: EntityQueryContextProvider;
}
/**
 * Defines an interface for a cache adapter flavor. An entity that uses a flavor will use the
 * specified adapter for caching.
 */
export interface CacheAdapterFlavorDefinition {
    cacheAdapterProvider: IEntityCacheAdapterProvider;
}
/**
 * Definition for constructing a companion for an entity. Defines the core set of objects
 * used to power the entity framework for a particular type of entity.
 */
export declare class EntityCompanionDefinition<TFields, TID extends NonNullable<TFields[TSelectedFields]>, TViewerContext extends ViewerContext, TEntity extends ReadonlyEntity<TFields, TID, TViewerContext, TSelectedFields>, TPrivacyPolicy extends EntityPrivacyPolicy<TFields, TID, TViewerContext, TEntity, TSelectedFields>, TSelectedFields extends keyof TFields = keyof TFields> {
    readonly entityClass: IEntityClass<TFields, TID, TViewerContext, TEntity, TPrivacyPolicy, TSelectedFields>;
    readonly entityConfiguration: EntityConfiguration<TFields>;
    readonly privacyPolicyClass: IPrivacyPolicyClass<TPrivacyPolicy>;
    readonly mutationValidators: () => EntityMutationValidator<TFields, TID, TViewerContext, TEntity, TSelectedFields>[];
    readonly mutationTriggers: () => EntityMutationTriggerConfiguration<TFields, TID, TViewerContext, TEntity, TSelectedFields>;
    readonly entitySelectedFields: TSelectedFields[];
    constructor({ entityClass, entityConfiguration, privacyPolicyClass, mutationValidators, mutationTriggers, entitySelectedFields, }: {
        /**
         * The concrete Entity class for which this is the definition.
         */
        entityClass: IEntityClass<TFields, TID, TViewerContext, TEntity, TPrivacyPolicy, TSelectedFields>;
        /**
         * The {@link EntityConfiguration} for this entity.
         */
        entityConfiguration: EntityConfiguration<TFields>;
        /**
         * The {@link EntityPrivacyPolicy} class for this entity.
         */
        privacyPolicyClass: IPrivacyPolicyClass<TPrivacyPolicy>;
        /**
         * An optional list of {@link EntityMutationValidator} for this entity.
         */
        mutationValidators?: () => EntityMutationValidator<TFields, TID, TViewerContext, TEntity, TSelectedFields>[];
        /**
         * An optional list of {@link EntityMutationTrigger} for this entity.
         */
        mutationTriggers?: () => EntityMutationTriggerConfiguration<TFields, TID, TViewerContext, TEntity, TSelectedFields>;
        /**
         * An optional subset of fields defined in the {@link EntityConfiguration} which belong to this entity.
         * For use when multiple types of entities are backed by a single table ({@link EntityConfiguration}) yet
         * only expose a subset of the fields.
         */
        entitySelectedFields?: TSelectedFields[];
    });
}
/**
 * An instance of the Entity framework.
 *
 * Required to create a {@link ViewerContext}, which is the application entry point
 * into the framework.
 *
 * Internally, this is a lazy entity companion factory that instantiates and caches one
 * {@link EntityCompanion} for each type of {@link Entity}.
 */
export default class EntityCompanionProvider {
    private metricsAdapter;
    private databaseAdapterFlavors;
    private cacheAdapterFlavors;
    private readonly companionMap;
    private readonly tableDataCoordinatorMap;
    /**
     * Instantiate an Entity framework.
     * @param metricsAdapter - An {@link IEntityMetricsAdapter} for collecting metrics on this instance
     * @param databaseAdapterFlavors - Database adapter configurations for this instance
     * @param cacheAdapterFlavors - Cache adapter configurations for this instance
     */
    constructor(metricsAdapter: IEntityMetricsAdapter, databaseAdapterFlavors: ReadonlyMap<DatabaseAdapterFlavor, DatabaseAdapterFlavorDefinition>, cacheAdapterFlavors: ReadonlyMap<CacheAdapterFlavor, CacheAdapterFlavorDefinition>);
    /**
     * Get the entity companion for specified entity. If not already computed and cached, the entity
     * companion is constructed using the configuration provided by the factory.
     *
     * @param entityClass - entity class to load
     * @param factory - entity companion factory
     */
    getCompanionForEntity<TFields, TID extends NonNullable<TFields[TSelectedFields]>, TViewerContext extends ViewerContext, TEntity extends ReadonlyEntity<TFields, TID, TViewerContext, TSelectedFields>, TPrivacyPolicy extends EntityPrivacyPolicy<TFields, TID, TViewerContext, TEntity, TSelectedFields>, TSelectedFields extends keyof TFields>(entityClass: IEntityClass<TFields, TID, TViewerContext, TEntity, TPrivacyPolicy, TSelectedFields>, entityCompanionDefinition: EntityCompanionDefinition<TFields, TID, TViewerContext, TEntity, TPrivacyPolicy, TSelectedFields>): EntityCompanion<TFields, TID, TViewerContext, TEntity, TPrivacyPolicy, TSelectedFields>;
    getQueryContextProviderForDatabaseAdaptorFlavor(databaseAdapterFlavor: DatabaseAdapterFlavor): EntityQueryContextProvider;
    private getTableDataCoordinatorForEntity;
}
