"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.EntityCompanionDefinition = void 0;
const invariant_1 = __importDefault(require("invariant"));
const EntityCompanion_1 = __importDefault(require("./EntityCompanion"));
const EntityTableDataCoordinator_1 = __importDefault(require("./internal/EntityTableDataCoordinator"));
const maps_1 = require("./utils/collections/maps");
/**
 * Definition for constructing a companion for an entity. Defines the core set of objects
 * used to power the entity framework for a particular type of entity.
 */
class EntityCompanionDefinition {
    constructor({ entityClass, entityConfiguration, privacyPolicyClass, mutationValidators = () => [], mutationTriggers = () => ({}), entitySelectedFields = Array.from(entityConfiguration.schema.keys()), }) {
        this.entityClass = entityClass;
        this.entityConfiguration = entityConfiguration;
        this.privacyPolicyClass = privacyPolicyClass;
        this.mutationValidators = mutationValidators;
        this.mutationTriggers = mutationTriggers;
        this.entitySelectedFields = entitySelectedFields;
    }
}
exports.EntityCompanionDefinition = EntityCompanionDefinition;
/**
 * An instance of the Entity framework.
 *
 * Required to create a {@link ViewerContext}, which is the application entry point
 * into the framework.
 *
 * Internally, this is a lazy entity companion factory that instantiates and caches one
 * {@link EntityCompanion} for each type of {@link Entity}.
 */
class EntityCompanionProvider {
    /**
     * Instantiate an Entity framework.
     * @param metricsAdapter - An {@link IEntityMetricsAdapter} for collecting metrics on this instance
     * @param databaseAdapterFlavors - Database adapter configurations for this instance
     * @param cacheAdapterFlavors - Cache adapter configurations for this instance
     */
    constructor(metricsAdapter, databaseAdapterFlavors, cacheAdapterFlavors) {
        this.metricsAdapter = metricsAdapter;
        this.databaseAdapterFlavors = databaseAdapterFlavors;
        this.cacheAdapterFlavors = cacheAdapterFlavors;
        this.companionMap = new Map();
        this.tableDataCoordinatorMap = new Map();
    }
    /**
     * Get the entity companion for specified entity. If not already computed and cached, the entity
     * companion is constructed using the configuration provided by the factory.
     *
     * @param entityClass - entity class to load
     * @param factory - entity companion factory
     */
    getCompanionForEntity(entityClass, entityCompanionDefinition) {
        const tableDataCoordinator = this.getTableDataCoordinatorForEntity(entityCompanionDefinition.entityConfiguration, entityClass.name);
        return maps_1.computeIfAbsent(this.companionMap, entityClass.name, () => {
            return new EntityCompanion_1.default(entityCompanionDefinition.entityClass, tableDataCoordinator, entityCompanionDefinition.privacyPolicyClass, entityCompanionDefinition.mutationValidators(), entityCompanionDefinition.mutationTriggers(), this.metricsAdapter);
        });
    }
    getQueryContextProviderForDatabaseAdaptorFlavor(databaseAdapterFlavor) {
        const entityDatabaseAdapterFlavor = this.databaseAdapterFlavors.get(databaseAdapterFlavor);
        invariant_1.default(entityDatabaseAdapterFlavor, `No database adaptor configuration found for flavor: ${databaseAdapterFlavor}`);
        return entityDatabaseAdapterFlavor.queryContextProvider;
    }
    getTableDataCoordinatorForEntity(entityConfiguration, entityClassName) {
        return maps_1.computeIfAbsent(this.tableDataCoordinatorMap, entityConfiguration.tableName, () => {
            const entityDatabaseAdapterFlavor = this.databaseAdapterFlavors.get(entityConfiguration.databaseAdapterFlavor);
            invariant_1.default(entityDatabaseAdapterFlavor, `No database adaptor configuration found for flavor: ${entityConfiguration.databaseAdapterFlavor}`);
            const entityCacheAdapterFlavor = this.cacheAdapterFlavors.get(entityConfiguration.cacheAdapterFlavor);
            invariant_1.default(entityCacheAdapterFlavor, `No cache adaptor configuration found for flavor: ${entityConfiguration.cacheAdapterFlavor}`);
            return new EntityTableDataCoordinator_1.default(entityConfiguration, entityDatabaseAdapterFlavor.adapterProvider, entityCacheAdapterFlavor.cacheAdapterProvider, entityDatabaseAdapterFlavor.queryContextProvider, this.metricsAdapter, entityClassName);
        });
    }
}
exports.default = EntityCompanionProvider;
//# sourceMappingURL=EntityCompanionProvider.js.map