"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.MaybeJSONArrayField = exports.JSONArrayField = exports.EnumField = exports.JSONObjectField = exports.StringArrayField = exports.FloatField = exports.IntField = exports.BooleanField = exports.DateField = exports.UUIDField = exports.StringField = exports.EntityFieldDefinition = exports.EntityEdgeDeletionBehavior = void 0;
const uuid_1 = require("uuid");
var EntityEdgeDeletionBehavior;
(function (EntityEdgeDeletionBehavior) {
    /**
     * Default. Invalidate the cache for all entities that reference the entity
     * being deleted through this field, and transitively run deletions on those entities.
     * This is most useful when the database itself expresses foreign
     * keys and cascading deletes or set nulls and the entity framework just needs to
     * be kept consistent with the state of the database.
     */
    EntityEdgeDeletionBehavior[EntityEdgeDeletionBehavior["CASCADE_DELETE_INVALIDATE_CACHE"] = 0] = "CASCADE_DELETE_INVALIDATE_CACHE";
    /**
     * Delete all entities that reference the entity being deleted through this field. This is very similar
     * to SQL `ON DELETE CASCADE` but is done in the Entity framework instead of at the underlying level.
     * This will also invalidate the cached referencing entities.
     */
    EntityEdgeDeletionBehavior[EntityEdgeDeletionBehavior["CASCADE_DELETE"] = 1] = "CASCADE_DELETE";
    /**
     * Set this field to null when the referenced entity is deleted. This is very similar
     * to SQL `ON DELETE SET NULL` but is done in the Entity framework instead of at the underlying level.
     * This will also invalidate the cached referencing entities.
     */
    EntityEdgeDeletionBehavior[EntityEdgeDeletionBehavior["SET_NULL"] = 2] = "SET_NULL";
})(EntityEdgeDeletionBehavior = exports.EntityEdgeDeletionBehavior || (exports.EntityEdgeDeletionBehavior = {}));
class EntityFieldDefinition {
    /**
     *
     * @param columnName - Column name in the database.
     * @param cache - Whether or not to cache loaded instances of the entity by this field. The column name is
     *              used to derive a cache key for the cache entry. If true, this column must be able uniquely
     *              identify the entity.
     */
    constructor({ columnName, cache = false, association, }) {
        this.columnName = columnName;
        this.cache = cache;
        this.association = association;
    }
    /**
     * Validates input value for a field of this type. Null and undefined are considered valid by default. This is used for things like:
     * - EntityLoader.loadByFieldValue - to ensure the value being loaded by is a valid value
     * - EntityMutator.setField - to ensure the value being set is a valid value
     */
    validateInputValue(value) {
        if (value === null || value === undefined) {
            return true;
        }
        return this.validateInputValueInternal(value);
    }
}
exports.EntityFieldDefinition = EntityFieldDefinition;
class StringField extends EntityFieldDefinition {
    validateInputValueInternal(value) {
        return typeof value === 'string';
    }
}
exports.StringField = StringField;
class UUIDField extends StringField {
    validateInputValueInternal(value) {
        return uuid_1.validate(value);
    }
}
exports.UUIDField = UUIDField;
class DateField extends EntityFieldDefinition {
    validateInputValueInternal(value) {
        return value instanceof Date;
    }
}
exports.DateField = DateField;
class BooleanField extends EntityFieldDefinition {
    validateInputValueInternal(value) {
        return typeof value === 'boolean';
    }
}
exports.BooleanField = BooleanField;
class IntField extends EntityFieldDefinition {
    validateInputValueInternal(value) {
        return typeof value === 'number' && Number.isInteger(value);
    }
}
exports.IntField = IntField;
class FloatField extends EntityFieldDefinition {
    validateInputValueInternal(value) {
        return typeof value === 'number';
    }
}
exports.FloatField = FloatField;
class StringArrayField extends EntityFieldDefinition {
    validateInputValueInternal(value) {
        return Array.isArray(value) && value.every((subValue) => typeof subValue === 'string');
    }
}
exports.StringArrayField = StringArrayField;
class JSONObjectField extends EntityFieldDefinition {
    validateInputValueInternal(value) {
        return typeof value === 'object' && !Array.isArray(value);
    }
}
exports.JSONObjectField = JSONObjectField;
class EnumField extends EntityFieldDefinition {
    validateInputValueInternal(value) {
        return typeof value === 'number' || typeof value === 'string';
    }
}
exports.EnumField = EnumField;
class JSONArrayField extends EntityFieldDefinition {
    validateInputValueInternal(value) {
        return Array.isArray(value);
    }
}
exports.JSONArrayField = JSONArrayField;
class MaybeJSONArrayField extends EntityFieldDefinition {
    validateInputValueInternal(_value) {
        return true;
    }
}
exports.MaybeJSONArrayField = MaybeJSONArrayField;
//# sourceMappingURL=EntityFields.js.map