import { Result } from '@expo/results';
import EnforcingEntityLoader from './EnforcingEntityLoader';
import { IEntityClass } from './Entity';
import EntityConfiguration from './EntityConfiguration';
import { FieldEqualityCondition, QuerySelectionModifiers } from './EntityDatabaseAdapter';
import EntityPrivacyPolicy from './EntityPrivacyPolicy';
import { EntityQueryContext } from './EntityQueryContext';
import ReadonlyEntity from './ReadonlyEntity';
import ViewerContext from './ViewerContext';
import EntityDataManager from './internal/EntityDataManager';
/**
 * The primary interface for loading entities. All normal loads are batched,
 * cached, and authorized against the entity's {@link EntityPrivacyPolicy}.
 */
export default class EntityLoader<TFields, TID extends NonNullable<TFields[TSelectedFields]>, TViewerContext extends ViewerContext, TEntity extends ReadonlyEntity<TFields, TID, TViewerContext, TSelectedFields>, TPrivacyPolicy extends EntityPrivacyPolicy<TFields, TID, TViewerContext, TEntity, TSelectedFields>, TSelectedFields extends keyof TFields> {
    private readonly viewerContext;
    private readonly queryContext;
    private readonly entityConfiguration;
    private readonly entityClass;
    private readonly privacyPolicy;
    private readonly dataManager;
    constructor(viewerContext: TViewerContext, queryContext: EntityQueryContext, entityConfiguration: EntityConfiguration<TFields>, entityClass: IEntityClass<TFields, TID, TViewerContext, TEntity, TPrivacyPolicy, TSelectedFields>, privacyPolicy: TPrivacyPolicy, dataManager: EntityDataManager<TFields>);
    /**
     * Enforcing view on this entity loader. All loads through this view are
     * guaranteed to be the values of successful results (or null for some loader methods),
     * and will throw otherwise.
     */
    enforcing(): EnforcingEntityLoader<TFields, TID, TViewerContext, TEntity, TPrivacyPolicy, TSelectedFields>;
    /**
     * Load many entities where fieldName is one of fieldValues.
     * @param fieldName - entity field being queried
     * @param fieldValues - fieldName field values being queried
     * @returns map from fieldValue to entity results that match the query for that fieldValue,
     *          where result errors can be UnauthorizedError
     */
    loadManyByFieldEqualingManyAsync<N extends keyof Pick<TFields, TSelectedFields>>(fieldName: N, fieldValues: readonly NonNullable<TFields[N]>[]): Promise<ReadonlyMap<NonNullable<TFields[N]>, readonly Result<TEntity>[]>>;
    /**
     * Load many entities where fieldName equals fieldValue.
     * @param fieldName - entity field being queried
     * @param fieldValue - fieldName field value being queried
     * @returns array of entity results that match the query for fieldValue, where result error can be UnauthorizedError
     */
    loadManyByFieldEqualingAsync<N extends keyof Pick<TFields, TSelectedFields>>(fieldName: N, fieldValue: NonNullable<TFields[N]>): Promise<readonly Result<TEntity>[]>;
    /**
     * Load an entity where fieldName equals fieldValue, or null if no entity exists.
     * @param uniqueFieldName - entity field being queried
     * @param fieldValue - uniqueFieldName field value being queried
     * @returns entity result where uniqueFieldName equals fieldValue, or null if no entity matches the condition.
     * @throws when multiple entities match the condition
     */
    loadByFieldEqualingAsync<N extends keyof Pick<TFields, TSelectedFields>>(uniqueFieldName: N, fieldValue: NonNullable<TFields[N]>): Promise<Result<TEntity> | null>;
    /**
     * Loads an entity by a specified ID.
     * @param id - ID of the entity
     * @returns entity result for matching ID, where result error can be UnauthorizedError or EntityNotFoundError.
     */
    loadByIDAsync(id: TID): Promise<Result<TEntity>>;
    /**
     * Load an entity by a specified ID, or return null if non-existent.
     * @param id - ID of the entity
     * @returns entity result for matching ID, or null if no entity exists for ID.
     */
    loadByIDNullableAsync(id: TID): Promise<Result<TEntity> | null>;
    /**
     * Loads many entities for a list of IDs.
     * @param viewerContext - viewer context of loading user
     * @param ids - IDs of the entities to load
     * @returns map from ID to corresponding entity result, where result error can be UnauthorizedError or EntityNotFoundError.
     */
    loadManyByIDsAsync(ids: readonly TID[]): Promise<ReadonlyMap<TID, Result<TEntity>>>;
    /**
     * Loads many entities matching the conjunction of WHERE clauses constructed from specified operands.
     * Entities loaded using this method are not batched or cached.
     *
     * @example
     * fieldEqualityOperands:
     * `[{fieldName: 'hello', fieldValue: 1}, {fieldName: 'world', fieldValues: [2, 3]}]`
     * Entities returned:
     * `WHERE hello = 1 AND world = ANY({2, 3})`
     *
     * @param fieldEqualityOperands - list of field equality WHERE clause operand specifications
     * @param querySelectionModifiers - limit, offset, and orderBy for the query
     * @returns array of entity results that match the query, where result error can be UnauthorizedError
     */
    loadManyByFieldEqualityConjunctionAsync<N extends keyof Pick<TFields, TSelectedFields>>(fieldEqualityOperands: FieldEqualityCondition<TFields, N>[], querySelectionModifiers?: QuerySelectionModifiers<TFields>): Promise<readonly Result<TEntity>[]>;
    /**
     * Loads many entities matching the raw WHERE clause. Corresponds to the knex `whereRaw` argument format.
     *
     * @remarks
     * Important notes:
     * - Fields in clause are database column names instead of transformed entity field names.
     * - Entities loaded using this method are not batched or cached.
     * - Not all database adapters implement the ability to execute this method of fetching entities.
     *
     * @example
     * rawWhereClause: `id = ?`
     * bindings: `[1]`
     * Entites returned `WHERE id = 1`
     *
     * {@link http://knexjs.org/#Builder-whereRaw}
     * {@link http://knexjs.org/#Raw-Bindings}
     *
     * @param rawWhereClause - parameterized SQL WHERE clause with positional binding placeholders or named binding placeholders
     * @param bindings - array of positional bindings or object of named bindings
     * @param querySelectionModifiers - limit, offset, and orderBy for the query
     * @returns array of entity results that match the query, where result error can be UnauthorizedError
     * @throws Error when rawWhereClause or bindings are invalid
     *
     * @deprecated prefer caching loaders
     */
    loadManyByRawWhereClauseAsync(rawWhereClause: string, bindings: any[] | object, querySelectionModifiers?: QuerySelectionModifiers<TFields>): Promise<readonly Result<TEntity>[]>;
    /**
     * Invalidate all caches for an entity's fields. Exposed primarily for internal use by {@link EntityMutator}.
     * @param objectFields - entity data object to be invalidated
     */
    invalidateFieldsAsync(objectFields: Readonly<TFields>): Promise<void>;
    /**
     * Invalidate all caches for an entity. One potential use case would be to keep the entity
     * framework in sync with changes made to data outside of the framework.
     * @param entity - entity to be invalidated
     */
    invalidateEntityAsync(entity: TEntity): Promise<void>;
    private tryConstructEntities;
    /**
     * Construct and authorize entities from fields map, returning error results for entities that fail
     * to construct or fail to authorize.
     *
     * @param map - map from an arbitrary key type to an array of entity field objects
     */
    constructAndAuthorizeEntitiesAsync<K>(map: ReadonlyMap<K, readonly Readonly<TFields>[]>): Promise<ReadonlyMap<K, readonly Result<TEntity>[]>>;
    private validateFieldValues;
}
