"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const results_1 = require("@expo/results");
const invariant_1 = __importDefault(require("invariant"));
const EnforcingEntityLoader_1 = __importDefault(require("./EnforcingEntityLoader"));
const EntityDatabaseAdapter_1 = require("./EntityDatabaseAdapter");
const EntityInvalidFieldValueError_1 = __importDefault(require("./errors/EntityInvalidFieldValueError"));
const EntityNotFoundError_1 = __importDefault(require("./errors/EntityNotFoundError"));
const maps_1 = require("./utils/collections/maps");
/**
 * The primary interface for loading entities. All normal loads are batched,
 * cached, and authorized against the entity's {@link EntityPrivacyPolicy}.
 */
class EntityLoader {
    constructor(viewerContext, queryContext, entityConfiguration, entityClass, privacyPolicy, dataManager) {
        this.viewerContext = viewerContext;
        this.queryContext = queryContext;
        this.entityConfiguration = entityConfiguration;
        this.entityClass = entityClass;
        this.privacyPolicy = privacyPolicy;
        this.dataManager = dataManager;
    }
    /**
     * Enforcing view on this entity loader. All loads through this view are
     * guaranteed to be the values of successful results (or null for some loader methods),
     * and will throw otherwise.
     */
    enforcing() {
        return new EnforcingEntityLoader_1.default(this);
    }
    /**
     * Load many entities where fieldName is one of fieldValues.
     * @param fieldName - entity field being queried
     * @param fieldValues - fieldName field values being queried
     * @returns map from fieldValue to entity results that match the query for that fieldValue,
     *          where result errors can be UnauthorizedError
     */
    async loadManyByFieldEqualingManyAsync(fieldName, fieldValues) {
        this.validateFieldValues(fieldName, fieldValues);
        const fieldValuesToFieldObjects = await this.dataManager.loadManyByFieldEqualingAsync(this.queryContext, fieldName, fieldValues);
        return await this.constructAndAuthorizeEntitiesAsync(fieldValuesToFieldObjects);
    }
    /**
     * Load many entities where fieldName equals fieldValue.
     * @param fieldName - entity field being queried
     * @param fieldValue - fieldName field value being queried
     * @returns array of entity results that match the query for fieldValue, where result error can be UnauthorizedError
     */
    async loadManyByFieldEqualingAsync(fieldName, fieldValue) {
        const entityResults = await this.loadManyByFieldEqualingManyAsync(fieldName, [fieldValue]);
        const entityResultsForFieldValue = entityResults.get(fieldValue);
        invariant_1.default(entityResultsForFieldValue !== undefined, `${fieldValue} should be guaranteed to be present in returned map of entities`);
        return entityResultsForFieldValue;
    }
    /**
     * Load an entity where fieldName equals fieldValue, or null if no entity exists.
     * @param uniqueFieldName - entity field being queried
     * @param fieldValue - uniqueFieldName field value being queried
     * @returns entity result where uniqueFieldName equals fieldValue, or null if no entity matches the condition.
     * @throws when multiple entities match the condition
     */
    async loadByFieldEqualingAsync(uniqueFieldName, fieldValue) {
        var _a;
        const entityResults = await this.loadManyByFieldEqualingAsync(uniqueFieldName, fieldValue);
        invariant_1.default(entityResults.length <= 1, `loadByFieldEqualing: Multiple entities of type ${this.entityClass.name} found for ${uniqueFieldName}=${fieldValue}`);
        return (_a = entityResults[0]) !== null && _a !== void 0 ? _a : null;
    }
    /**
     * Loads an entity by a specified ID.
     * @param id - ID of the entity
     * @returns entity result for matching ID, where result error can be UnauthorizedError or EntityNotFoundError.
     */
    async loadByIDAsync(id) {
        const entityResults = await this.loadManyByIDsAsync([id]);
        const entityResult = entityResults.get(id);
        if (entityResult === undefined) {
            return results_1.result(new EntityNotFoundError_1.default(this.entityClass, this.entityConfiguration.idField, id));
        }
        return entityResult;
    }
    /**
     * Load an entity by a specified ID, or return null if non-existent.
     * @param id - ID of the entity
     * @returns entity result for matching ID, or null if no entity exists for ID.
     */
    async loadByIDNullableAsync(id) {
        return await this.loadByFieldEqualingAsync(this.entityConfiguration.idField, id);
    }
    /**
     * Loads many entities for a list of IDs.
     * @param viewerContext - viewer context of loading user
     * @param ids - IDs of the entities to load
     * @returns map from ID to corresponding entity result, where result error can be UnauthorizedError or EntityNotFoundError.
     */
    async loadManyByIDsAsync(ids) {
        const entityResults = (await this.loadManyByFieldEqualingManyAsync(this.entityConfiguration.idField, ids));
        return maps_1.mapMap(entityResults, (entityResultsForId, id) => {
            const entityResult = entityResultsForId[0];
            return (entityResult !== null && entityResult !== void 0 ? entityResult : results_1.result(new EntityNotFoundError_1.default(this.entityClass, this.entityConfiguration.idField, id)));
        });
    }
    /**
     * Loads many entities matching the conjunction of WHERE clauses constructed from specified operands.
     * Entities loaded using this method are not batched or cached.
     *
     * @example
     * fieldEqualityOperands:
     * `[{fieldName: 'hello', fieldValue: 1}, {fieldName: 'world', fieldValues: [2, 3]}]`
     * Entities returned:
     * `WHERE hello = 1 AND world = ANY({2, 3})`
     *
     * @param fieldEqualityOperands - list of field equality WHERE clause operand specifications
     * @param querySelectionModifiers - limit, offset, and orderBy for the query
     * @returns array of entity results that match the query, where result error can be UnauthorizedError
     */
    async loadManyByFieldEqualityConjunctionAsync(fieldEqualityOperands, querySelectionModifiers = {}) {
        for (const fieldEqualityOperand of fieldEqualityOperands) {
            const fieldValues = EntityDatabaseAdapter_1.isSingleValueFieldEqualityCondition(fieldEqualityOperand)
                ? [fieldEqualityOperand.fieldValue]
                : fieldEqualityOperand.fieldValues;
            this.validateFieldValues(fieldEqualityOperand.fieldName, fieldValues);
        }
        const fieldObjects = await this.dataManager.loadManyByFieldEqualityConjunctionAsync(this.queryContext, fieldEqualityOperands, querySelectionModifiers);
        const uncheckedEntityResults = this.tryConstructEntities(fieldObjects);
        return await Promise.all(uncheckedEntityResults.map(async (uncheckedEntityResult) => {
            if (!uncheckedEntityResult.ok) {
                return uncheckedEntityResult;
            }
            return await results_1.asyncResult(this.privacyPolicy.authorizeReadAsync(this.viewerContext, this.queryContext, uncheckedEntityResult.value));
        }));
    }
    /**
     * Loads many entities matching the raw WHERE clause. Corresponds to the knex `whereRaw` argument format.
     *
     * @remarks
     * Important notes:
     * - Fields in clause are database column names instead of transformed entity field names.
     * - Entities loaded using this method are not batched or cached.
     * - Not all database adapters implement the ability to execute this method of fetching entities.
     *
     * @example
     * rawWhereClause: `id = ?`
     * bindings: `[1]`
     * Entites returned `WHERE id = 1`
     *
     * {@link http://knexjs.org/#Builder-whereRaw}
     * {@link http://knexjs.org/#Raw-Bindings}
     *
     * @param rawWhereClause - parameterized SQL WHERE clause with positional binding placeholders or named binding placeholders
     * @param bindings - array of positional bindings or object of named bindings
     * @param querySelectionModifiers - limit, offset, and orderBy for the query
     * @returns array of entity results that match the query, where result error can be UnauthorizedError
     * @throws Error when rawWhereClause or bindings are invalid
     *
     * @deprecated prefer caching loaders
     */
    async loadManyByRawWhereClauseAsync(rawWhereClause, bindings, querySelectionModifiers = {}) {
        const fieldObjects = await this.dataManager.loadManyByRawWhereClauseAsync(this.queryContext, rawWhereClause, bindings, querySelectionModifiers);
        const uncheckedEntityResults = this.tryConstructEntities(fieldObjects);
        return await Promise.all(uncheckedEntityResults.map(async (uncheckedEntityResult) => {
            if (!uncheckedEntityResult.ok) {
                return uncheckedEntityResult;
            }
            return await results_1.asyncResult(this.privacyPolicy.authorizeReadAsync(this.viewerContext, this.queryContext, uncheckedEntityResult.value));
        }));
    }
    /**
     * Invalidate all caches for an entity's fields. Exposed primarily for internal use by {@link EntityMutator}.
     * @param objectFields - entity data object to be invalidated
     */
    async invalidateFieldsAsync(objectFields) {
        await this.dataManager.invalidateObjectFieldsAsync(objectFields);
    }
    /**
     * Invalidate all caches for an entity. One potential use case would be to keep the entity
     * framework in sync with changes made to data outside of the framework.
     * @param entity - entity to be invalidated
     */
    async invalidateEntityAsync(entity) {
        await this.invalidateFieldsAsync(entity.getAllDatabaseFields());
    }
    tryConstructEntities(fieldsObjects) {
        return fieldsObjects.map((fieldsObject) => {
            try {
                return results_1.result(new this.entityClass(this.viewerContext, fieldsObject));
            }
            catch (e) {
                return results_1.result(e);
            }
        });
    }
    /**
     * Construct and authorize entities from fields map, returning error results for entities that fail
     * to construct or fail to authorize.
     *
     * @param map - map from an arbitrary key type to an array of entity field objects
     */
    async constructAndAuthorizeEntitiesAsync(map) {
        const uncheckedEntityResultsMap = maps_1.mapMap(map, (fieldObjects) => this.tryConstructEntities(fieldObjects));
        return await maps_1.mapMapAsync(uncheckedEntityResultsMap, async (uncheckedEntityResults) => {
            return await Promise.all(uncheckedEntityResults.map(async (uncheckedEntityResult) => {
                if (!uncheckedEntityResult.ok) {
                    return uncheckedEntityResult;
                }
                return await results_1.asyncResult(this.privacyPolicy.authorizeReadAsync(this.viewerContext, this.queryContext, uncheckedEntityResult.value));
            }));
        });
    }
    validateFieldValues(fieldName, fieldValues) {
        const fieldDefinition = this.entityConfiguration.schema.get(fieldName);
        invariant_1.default(fieldDefinition, `must have field definition for field = ${fieldName}`);
        for (const fieldValue of fieldValues) {
            const isInputValid = fieldDefinition.validateInputValue(fieldValue);
            if (!isInputValid) {
                throw new EntityInvalidFieldValueError_1.default(this.entityClass, fieldName, fieldValue);
            }
        }
    }
}
exports.default = EntityLoader;
//# sourceMappingURL=EntityLoader.js.map