"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DeleteMutator = exports.UpdateMutator = exports.CreateMutator = exports.EntityMutationType = void 0;
const results_1 = require("@expo/results");
const invariant_1 = __importDefault(require("invariant"));
const EntityFields_1 = require("./EntityFields");
const EntityInvalidFieldValueError_1 = __importDefault(require("./errors/EntityInvalidFieldValueError"));
const EntityMetricsUtils_1 = require("./metrics/EntityMetricsUtils");
const IEntityMetricsAdapter_1 = require("./metrics/IEntityMetricsAdapter");
const maps_1 = require("./utils/collections/maps");
var EntityMutationType;
(function (EntityMutationType) {
    EntityMutationType[EntityMutationType["CREATE"] = 0] = "CREATE";
    EntityMutationType[EntityMutationType["UPDATE"] = 1] = "UPDATE";
    EntityMutationType[EntityMutationType["DELETE"] = 2] = "DELETE";
})(EntityMutationType = exports.EntityMutationType || (exports.EntityMutationType = {}));
class BaseMutator {
    constructor(viewerContext, queryContext, entityConfiguration, entityClass, privacyPolicy, mutationValidators, mutationTriggers, entityLoaderFactory, databaseAdapter, metricsAdapter) {
        this.viewerContext = viewerContext;
        this.queryContext = queryContext;
        this.entityConfiguration = entityConfiguration;
        this.entityClass = entityClass;
        this.privacyPolicy = privacyPolicy;
        this.mutationValidators = mutationValidators;
        this.mutationTriggers = mutationTriggers;
        this.entityLoaderFactory = entityLoaderFactory;
        this.databaseAdapter = databaseAdapter;
        this.metricsAdapter = metricsAdapter;
    }
    validateFields(fields) {
        for (const fieldName in fields) {
            const fieldValue = fields[fieldName];
            const fieldDefinition = this.entityConfiguration.schema.get(fieldName);
            invariant_1.default(fieldDefinition, `must have field definition for field = ${fieldName}`);
            const isInputValid = fieldDefinition.validateInputValue(fieldValue);
            if (!isInputValid) {
                throw new EntityInvalidFieldValueError_1.default(this.entityClass, fieldName, fieldValue);
            }
        }
    }
    async executeMutationTriggersOrValidatorsAsync(triggersOrValidators, queryContext, entity, mutationInfo) {
        if (!triggersOrValidators) {
            return;
        }
        await Promise.all(triggersOrValidators.map((triggerOrValidator) => triggerOrValidator.executeAsync(this.viewerContext, queryContext, entity, mutationInfo)));
    }
    async executeNonTransactionalMutationTriggersOrValidatorsAsync(triggers, entity, mutationInfo) {
        if (!triggers) {
            return;
        }
        await Promise.all(triggers.map((trigger) => trigger.executeAsync(this.viewerContext, entity, mutationInfo)));
    }
}
/**
 * Mutator for creating a new entity.
 */
class CreateMutator extends BaseMutator {
    constructor() {
        super(...arguments);
        this.fieldsForEntity = {};
    }
    /**
     * Set the value for entity field.
     * @param fieldName - entity field being updated
     * @param value - value for entity field
     */
    setField(fieldName, value) {
        this.fieldsForEntity[fieldName] = value;
        return this;
    }
    /**
     * Commit the new entity after authorizing against creation privacy rules. Invalidates all caches for
     * queries that would return new entity and caches the new entity if not in a transactional query context.
     * @returns authorized, cached, newly-created entity result, where result error can be UnauthorizedError
     */
    async createAsync() {
        return await EntityMetricsUtils_1.timeAndLogMutationEventAsync(this.metricsAdapter, IEntityMetricsAdapter_1.EntityMetricsMutationType.CREATE, this.entityClass.name)(this.createInTransactionAsync());
    }
    /**
     * Convenience method that returns the new entity or throws upon create failure.
     */
    async enforceCreateAsync() {
        return await results_1.enforceAsyncResult(this.createAsync());
    }
    async createInTransactionAsync() {
        return await this.queryContext.runInTransactionIfNotInTransactionAsync((innerQueryContext) => this.createInternalAsync(innerQueryContext));
    }
    async createInternalAsync(queryContext) {
        this.validateFields(this.fieldsForEntity);
        const temporaryEntityForPrivacyCheck = new this.entityClass(this.viewerContext, {
            [this.entityConfiguration.idField]: '00000000-0000-0000-0000-000000000000',
            ...this.fieldsForEntity,
        });
        const authorizeCreateResult = await results_1.asyncResult(this.privacyPolicy.authorizeCreateAsync(this.viewerContext, queryContext, temporaryEntityForPrivacyCheck));
        if (!authorizeCreateResult.ok) {
            return authorizeCreateResult;
        }
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationValidators, queryContext, temporaryEntityForPrivacyCheck, { type: EntityMutationType.CREATE });
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.beforeAll, queryContext, temporaryEntityForPrivacyCheck, { type: EntityMutationType.CREATE });
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.beforeCreate, queryContext, temporaryEntityForPrivacyCheck, { type: EntityMutationType.CREATE });
        const insertResult = await this.databaseAdapter.insertAsync(queryContext, this.fieldsForEntity);
        const entityLoader = this.entityLoaderFactory.forLoad(this.viewerContext, queryContext);
        queryContext.appendPostCommitCallback(entityLoader.invalidateFieldsAsync.bind(entityLoader, insertResult));
        const unauthorizedEntityAfterInsert = new this.entityClass(this.viewerContext, insertResult);
        const newEntity = await entityLoader
            .enforcing()
            .loadByIDAsync(unauthorizedEntityAfterInsert.getID());
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.afterCreate, queryContext, newEntity, { type: EntityMutationType.CREATE });
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.afterAll, queryContext, newEntity, { type: EntityMutationType.CREATE });
        queryContext.appendPostCommitCallback(this.executeNonTransactionalMutationTriggersOrValidatorsAsync.bind(this, this.mutationTriggers.afterCommit, newEntity, { type: EntityMutationType.CREATE }));
        return results_1.result(newEntity);
    }
}
exports.CreateMutator = CreateMutator;
/**
 * Mutator for updating an existing entity.
 */
class UpdateMutator extends BaseMutator {
    constructor(viewerContext, queryContext, entityConfiguration, entityClass, privacyPolicy, mutationValidators, mutationTriggers, entityLoaderFactory, databaseAdapter, metricsAdapter, originalEntity) {
        super(viewerContext, queryContext, entityConfiguration, entityClass, privacyPolicy, mutationValidators, mutationTriggers, entityLoaderFactory, databaseAdapter, metricsAdapter);
        this.updatedFields = {};
        this.originalEntity = originalEntity;
        this.fieldsForEntity = { ...originalEntity.getAllDatabaseFields() };
    }
    /**
     * Set the value for entity field.
     * @param fieldName - entity field being updated
     * @param value - value for entity field
     */
    setField(fieldName, value) {
        this.fieldsForEntity[fieldName] = value;
        this.updatedFields[fieldName] = value;
        return this;
    }
    /**
     * Commit the changes to the entity after authorizing against update privacy rules.
     * Invalidates all caches for pre-update entity and caches the updated entity if not in a
     * transactional query context.
     * @returns authorized updated entity result, where result error can be UnauthorizedError
     */
    async updateAsync() {
        return await EntityMetricsUtils_1.timeAndLogMutationEventAsync(this.metricsAdapter, IEntityMetricsAdapter_1.EntityMetricsMutationType.UPDATE, this.entityClass.name)(this.updateInTransactionAsync());
    }
    /**
     * Convenience method that returns the updated entity or throws upon update failure.
     */
    async enforceUpdateAsync() {
        return await results_1.enforceAsyncResult(this.updateAsync());
    }
    async updateInTransactionAsync() {
        return await this.queryContext.runInTransactionIfNotInTransactionAsync((innerQueryContext) => this.updateInternalAsync(innerQueryContext));
    }
    async updateInternalAsync(queryContext) {
        this.validateFields(this.updatedFields);
        const entityAboutToBeUpdated = new this.entityClass(this.viewerContext, this.fieldsForEntity);
        const authorizeUpdateResult = await results_1.asyncResult(this.privacyPolicy.authorizeUpdateAsync(this.viewerContext, queryContext, entityAboutToBeUpdated));
        if (!authorizeUpdateResult.ok) {
            return authorizeUpdateResult;
        }
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationValidators, queryContext, entityAboutToBeUpdated, { type: EntityMutationType.UPDATE, previousValue: this.originalEntity });
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.beforeAll, queryContext, entityAboutToBeUpdated, { type: EntityMutationType.UPDATE, previousValue: this.originalEntity });
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.beforeUpdate, queryContext, entityAboutToBeUpdated, { type: EntityMutationType.UPDATE, previousValue: this.originalEntity });
        const updateResult = await this.databaseAdapter.updateAsync(queryContext, this.entityConfiguration.idField, entityAboutToBeUpdated.getID(), this.updatedFields);
        const entityLoader = this.entityLoaderFactory.forLoad(this.viewerContext, queryContext);
        queryContext.appendPostCommitCallback(entityLoader.invalidateFieldsAsync.bind(entityLoader, this.originalEntity.getAllDatabaseFields()));
        queryContext.appendPostCommitCallback(entityLoader.invalidateFieldsAsync.bind(entityLoader, this.fieldsForEntity));
        const unauthorizedEntityAfterUpdate = new this.entityClass(this.viewerContext, updateResult);
        const updatedEntity = await entityLoader
            .enforcing()
            .loadByIDAsync(unauthorizedEntityAfterUpdate.getID());
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.afterUpdate, queryContext, updatedEntity, { type: EntityMutationType.UPDATE, previousValue: this.originalEntity });
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.afterAll, queryContext, updatedEntity, { type: EntityMutationType.UPDATE, previousValue: this.originalEntity });
        queryContext.appendPostCommitCallback(this.executeNonTransactionalMutationTriggersOrValidatorsAsync.bind(this, this.mutationTriggers.afterCommit, updatedEntity, { type: EntityMutationType.UPDATE, previousValue: this.originalEntity }));
        return results_1.result(updatedEntity);
    }
}
exports.UpdateMutator = UpdateMutator;
/**
 * Mutator for deleting an existing entity.
 */
class DeleteMutator extends BaseMutator {
    constructor(viewerContext, queryContext, entityConfiguration, entityClass, privacyPolicy, mutationValidators, mutationTriggers, entityLoaderFactory, databaseAdapter, metricsAdapter, entity) {
        super(viewerContext, queryContext, entityConfiguration, entityClass, privacyPolicy, mutationValidators, mutationTriggers, entityLoaderFactory, databaseAdapter, metricsAdapter);
        this.entity = entity;
    }
    /**
     * Delete the entity after authorizing against delete privacy rules. The entity is invalidated in all caches.
     * @returns void result, where result error can be UnauthorizedError
     */
    async deleteAsync() {
        return await EntityMetricsUtils_1.timeAndLogMutationEventAsync(this.metricsAdapter, IEntityMetricsAdapter_1.EntityMetricsMutationType.DELETE, this.entityClass.name)(this.deleteInTransactionAsync());
    }
    /**
     * Convenience method that throws upon delete failure.
     */
    async enforceDeleteAsync() {
        return await results_1.enforceAsyncResult(this.deleteAsync());
    }
    async deleteInTransactionAsync(processedEntityIdentifiersFromTransitiveDeletions = new Set(), skipDatabaseDeletion = false) {
        return await this.queryContext.runInTransactionIfNotInTransactionAsync((innerQueryContext) => this.deleteInternalAsync(innerQueryContext, processedEntityIdentifiersFromTransitiveDeletions, skipDatabaseDeletion));
    }
    async deleteInternalAsync(queryContext, processedEntityIdentifiersFromTransitiveDeletions, skipDatabaseDeletion) {
        const authorizeDeleteResult = await results_1.asyncResult(this.privacyPolicy.authorizeDeleteAsync(this.viewerContext, queryContext, this.entity));
        if (!authorizeDeleteResult.ok) {
            return authorizeDeleteResult;
        }
        await this.processEntityDeletionForInboundEdgesAsync(this.entity, queryContext, processedEntityIdentifiersFromTransitiveDeletions);
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.beforeAll, queryContext, this.entity, { type: EntityMutationType.DELETE });
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.beforeDelete, queryContext, this.entity, { type: EntityMutationType.DELETE });
        if (!skipDatabaseDeletion) {
            await this.databaseAdapter.deleteAsync(queryContext, this.entityConfiguration.idField, this.entity.getID());
        }
        const entityLoader = this.entityLoaderFactory.forLoad(this.viewerContext, queryContext);
        queryContext.appendPostCommitCallback(entityLoader.invalidateFieldsAsync.bind(entityLoader, this.entity.getAllDatabaseFields()));
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.afterDelete, queryContext, this.entity, { type: EntityMutationType.DELETE });
        await this.executeMutationTriggersOrValidatorsAsync(this.mutationTriggers.afterAll, queryContext, this.entity, { type: EntityMutationType.DELETE });
        queryContext.appendPostCommitCallback(this.executeNonTransactionalMutationTriggersOrValidatorsAsync.bind(this, this.mutationTriggers.afterCommit, this.entity, { type: EntityMutationType.DELETE }));
        return results_1.result();
    }
    /**
     * Finds all entities referencing the specified entity and either deletes them, nullifies
     * their references to the specified entity, or invalidates the cache depending on the
     * {@link OnDeleteBehavior} of the field referencing the specified entity.
     *
     * @remarks
     * This works by doing reverse fan-out queries:
     * 1. Load all entity configurations of entity types that reference this type of entity
     * 2. For each entity configuration, find all fields that contain edges to this type of entity
     * 3. For each edge field, load all entities with an edge from target entity to this entity via that field
     * 4. Perform desired OnDeleteBehavior for entities
     *
     * @param entity - entity to find all references to
     */
    async processEntityDeletionForInboundEdgesAsync(entity, queryContext, processedEntityIdentifiers) {
        // prevent infinite reference cycles by keeping track of entities already processed
        if (processedEntityIdentifiers.has(entity.getUniqueIdentifier())) {
            return;
        }
        processedEntityIdentifiers.add(entity.getUniqueIdentifier());
        const companionDefinition = entity.constructor.getCompanionDefinition();
        const entityConfiguration = companionDefinition.entityConfiguration;
        const inboundEdges = entityConfiguration.getInboundEdges();
        await Promise.all(inboundEdges.map(async (entityClass) => {
            return await maps_1.mapMapAsync(entityClass.getCompanionDefinition().entityConfiguration.schema, async (fieldDefinition, fieldName) => {
                const association = fieldDefinition.association;
                if (!association) {
                    return;
                }
                const associatedConfiguration = association
                    .getAssociatedEntityClass()
                    .getCompanionDefinition().entityConfiguration;
                if (associatedConfiguration !== entityConfiguration) {
                    return;
                }
                const associatedEntityLookupByField = association.associatedEntityLookupByField;
                const loaderFactory = entity
                    .getViewerContext()
                    .getViewerScopedEntityCompanionForClass(entityClass)
                    .getLoaderFactory();
                const mutatorFactory = entity
                    .getViewerContext()
                    .getViewerScopedEntityCompanionForClass(entityClass)
                    .getMutatorFactory();
                let inboundReferenceEntities;
                if (associatedEntityLookupByField) {
                    inboundReferenceEntities = await loaderFactory
                        .forLoad(queryContext)
                        .enforcing()
                        .loadManyByFieldEqualingAsync(fieldName, entity.getField(associatedEntityLookupByField));
                }
                else {
                    inboundReferenceEntities = await loaderFactory
                        .forLoad(queryContext)
                        .enforcing()
                        .loadManyByFieldEqualingAsync(fieldName, entity.getID());
                }
                switch (association.edgeDeletionBehavior) {
                    case undefined:
                    case EntityFields_1.EntityEdgeDeletionBehavior.CASCADE_DELETE_INVALIDATE_CACHE: {
                        await Promise.all(inboundReferenceEntities.map((inboundReferenceEntity) => mutatorFactory
                            .forDelete(inboundReferenceEntity, queryContext)
                            .deleteInTransactionAsync(processedEntityIdentifiers, 
                        /* skipDatabaseDeletion */ true // deletion is handled by DB
                        )));
                        break;
                    }
                    case EntityFields_1.EntityEdgeDeletionBehavior.SET_NULL: {
                        await Promise.all(inboundReferenceEntities.map((inboundReferenceEntity) => mutatorFactory
                            .forUpdate(inboundReferenceEntity, queryContext)
                            .setField(fieldName, null)
                            .enforceUpdateAsync()));
                        break;
                    }
                    case EntityFields_1.EntityEdgeDeletionBehavior.CASCADE_DELETE: {
                        await Promise.all(inboundReferenceEntities.map((inboundReferenceEntity) => mutatorFactory
                            .forDelete(inboundReferenceEntity, queryContext)
                            .deleteInTransactionAsync(processedEntityIdentifiers, 
                        /* skipDatabaseDeletion */ false)));
                    }
                }
            });
        }));
    }
}
exports.DeleteMutator = DeleteMutator;
//# sourceMappingURL=EntityMutator.js.map