import { EntityQueryContext } from './EntityQueryContext';
import ReadonlyEntity from './ReadonlyEntity';
import ViewerContext from './ViewerContext';
import EntityNotAuthorizedError from './errors/EntityNotAuthorizedError';
import PrivacyPolicyRule from './rules/PrivacyPolicyRule';
export declare enum EntityPrivacyPolicyEvaluationMode {
    ENFORCE = 0,
    DRY_RUN = 1,
    ENFORCE_AND_LOG = 2
}
export declare type EntityPrivacyPolicyEvaluator<TFields, TID extends NonNullable<TFields[TSelectedFields]>, TViewerContext extends ViewerContext, TEntity extends ReadonlyEntity<TFields, TID, TViewerContext, TSelectedFields>, TSelectedFields extends keyof TFields = keyof TFields> = {
    mode: EntityPrivacyPolicyEvaluationMode.ENFORCE;
} | {
    mode: EntityPrivacyPolicyEvaluationMode.DRY_RUN;
    denyHandler: (error: EntityNotAuthorizedError<TFields, TID, TViewerContext, TEntity, TSelectedFields>) => void;
} | {
    mode: EntityPrivacyPolicyEvaluationMode.ENFORCE_AND_LOG;
    denyHandler: (error: EntityNotAuthorizedError<TFields, TID, TViewerContext, TEntity, TSelectedFields>) => void;
};
export declare enum EntityAuthorizationAction {
    CREATE = 0,
    READ = 1,
    UPDATE = 2,
    DELETE = 3
}
/**
 * Privacy policy for an entity.
 *
 * @remarks
 *
 * A privacy policy declares lists of {@link PrivacyPolicyRule} for create, read, update, and delete actions
 * for an entity and provides logic for authorizing an entity against rules.
 *
 * Evaluation of a list of rules is performed according the following example. This allows constructing of
 * complex yet testable permissioning logic for an entity.
 *
 * @example
 *
 * ```
 * foreach rule in rules:
 *   return authorized if rule allows
 *   return not authorized if rule denies
 *   continue to next rule if rule skips
 * return not authorized if all rules skip
 * ```
 */
export default abstract class EntityPrivacyPolicy<TFields, TID extends NonNullable<TFields[TSelectedFields]>, TViewerContext extends ViewerContext, TEntity extends ReadonlyEntity<TFields, TID, TViewerContext, TSelectedFields>, TSelectedFields extends keyof TFields = keyof TFields> {
    protected readonly createRules: readonly PrivacyPolicyRule<TFields, TID, TViewerContext, TEntity, TSelectedFields>[];
    protected readonly readRules: readonly PrivacyPolicyRule<TFields, TID, TViewerContext, TEntity, TSelectedFields>[];
    protected readonly updateRules: readonly PrivacyPolicyRule<TFields, TID, TViewerContext, TEntity, TSelectedFields>[];
    protected readonly deleteRules: readonly PrivacyPolicyRule<TFields, TID, TViewerContext, TEntity, TSelectedFields>[];
    /**
     * Get the privacy policy evaluation mode and deny handler for this policy.
     * Defaults to normal enforcing policy.
     *
     * @remarks
     *
     * Override to enable dry run evaluation of the policy.
     */
    protected getPrivacyPolicyEvaluator(_viewerContext: TViewerContext): EntityPrivacyPolicyEvaluator<TFields, TID, TViewerContext, TEntity, TSelectedFields>;
    /**
     * Authorize an entity against creation policy.
     * @param viewerContext - viewer context of user creating the entity
     * @param queryContext - query context in which to perform the create authorization
     * @param entity - entity to authorize
     * @returns entity if authorized
     * @throws {@link EntityNotAuthorizedError} when not authorized
     */
    authorizeCreateAsync(viewerContext: TViewerContext, queryContext: EntityQueryContext, entity: TEntity): Promise<TEntity>;
    /**
     * Authorize an entity against read policy.
     * @param viewerContext - viewer context of user reading the entity
     * @param queryContext - query context in which to perform the read authorization
     * @param entity - entity to authorize
     * @returns entity if authorized
     * @throws {@link EntityNotAuthorizedError} when not authorized
     */
    authorizeReadAsync(viewerContext: TViewerContext, queryContext: EntityQueryContext, entity: TEntity): Promise<TEntity>;
    /**
     * Authorize an entity against update policy.
     * @param viewerContext - viewer context of user updating the entity
     * @param queryContext - query context in which to perform the update authorization
     * @param entity - entity to authorize
     * @returns entity if authorized
     * @throws {@link EntityNotAuthorizedError} when not authorized
     */
    authorizeUpdateAsync(viewerContext: TViewerContext, queryContext: EntityQueryContext, entity: TEntity): Promise<TEntity>;
    /**
     * Authorize an entity against deletion policy.
     * @param viewerContext - viewer context of user deleting the entity
     * @param queryContext - query context in which to perform the delete authorization
     * @param entity - entity to authorize
     * @returns entity if authorized
     * @throws {@link EntityNotAuthorizedError} when not authorized
     */
    authorizeDeleteAsync(viewerContext: TViewerContext, queryContext: EntityQueryContext, entity: TEntity): Promise<TEntity>;
    private authorizeForRulesetAsync;
    private authorizeForRulesetInnerAsync;
}
