"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.EntityAuthorizationAction = exports.EntityPrivacyPolicyEvaluationMode = void 0;
const EntityNotAuthorizedError_1 = __importDefault(require("./errors/EntityNotAuthorizedError"));
const PrivacyPolicyRule_1 = require("./rules/PrivacyPolicyRule");
var EntityPrivacyPolicyEvaluationMode;
(function (EntityPrivacyPolicyEvaluationMode) {
    EntityPrivacyPolicyEvaluationMode[EntityPrivacyPolicyEvaluationMode["ENFORCE"] = 0] = "ENFORCE";
    EntityPrivacyPolicyEvaluationMode[EntityPrivacyPolicyEvaluationMode["DRY_RUN"] = 1] = "DRY_RUN";
    EntityPrivacyPolicyEvaluationMode[EntityPrivacyPolicyEvaluationMode["ENFORCE_AND_LOG"] = 2] = "ENFORCE_AND_LOG";
})(EntityPrivacyPolicyEvaluationMode = exports.EntityPrivacyPolicyEvaluationMode || (exports.EntityPrivacyPolicyEvaluationMode = {}));
var EntityAuthorizationAction;
(function (EntityAuthorizationAction) {
    EntityAuthorizationAction[EntityAuthorizationAction["CREATE"] = 0] = "CREATE";
    EntityAuthorizationAction[EntityAuthorizationAction["READ"] = 1] = "READ";
    EntityAuthorizationAction[EntityAuthorizationAction["UPDATE"] = 2] = "UPDATE";
    EntityAuthorizationAction[EntityAuthorizationAction["DELETE"] = 3] = "DELETE";
})(EntityAuthorizationAction = exports.EntityAuthorizationAction || (exports.EntityAuthorizationAction = {}));
/**
 * Privacy policy for an entity.
 *
 * @remarks
 *
 * A privacy policy declares lists of {@link PrivacyPolicyRule} for create, read, update, and delete actions
 * for an entity and provides logic for authorizing an entity against rules.
 *
 * Evaluation of a list of rules is performed according the following example. This allows constructing of
 * complex yet testable permissioning logic for an entity.
 *
 * @example
 *
 * ```
 * foreach rule in rules:
 *   return authorized if rule allows
 *   return not authorized if rule denies
 *   continue to next rule if rule skips
 * return not authorized if all rules skip
 * ```
 */
class EntityPrivacyPolicy {
    constructor() {
        this.createRules = [];
        this.readRules = [];
        this.updateRules = [];
        this.deleteRules = [];
    }
    /**
     * Get the privacy policy evaluation mode and deny handler for this policy.
     * Defaults to normal enforcing policy.
     *
     * @remarks
     *
     * Override to enable dry run evaluation of the policy.
     */
    getPrivacyPolicyEvaluator(_viewerContext) {
        return {
            mode: EntityPrivacyPolicyEvaluationMode.ENFORCE,
        };
    }
    /**
     * Authorize an entity against creation policy.
     * @param viewerContext - viewer context of user creating the entity
     * @param queryContext - query context in which to perform the create authorization
     * @param entity - entity to authorize
     * @returns entity if authorized
     * @throws {@link EntityNotAuthorizedError} when not authorized
     */
    async authorizeCreateAsync(viewerContext, queryContext, entity) {
        return await this.authorizeForRulesetAsync(this.createRules, viewerContext, queryContext, entity, EntityAuthorizationAction.CREATE);
    }
    /**
     * Authorize an entity against read policy.
     * @param viewerContext - viewer context of user reading the entity
     * @param queryContext - query context in which to perform the read authorization
     * @param entity - entity to authorize
     * @returns entity if authorized
     * @throws {@link EntityNotAuthorizedError} when not authorized
     */
    async authorizeReadAsync(viewerContext, queryContext, entity) {
        return await this.authorizeForRulesetAsync(this.readRules, viewerContext, queryContext, entity, EntityAuthorizationAction.READ);
    }
    /**
     * Authorize an entity against update policy.
     * @param viewerContext - viewer context of user updating the entity
     * @param queryContext - query context in which to perform the update authorization
     * @param entity - entity to authorize
     * @returns entity if authorized
     * @throws {@link EntityNotAuthorizedError} when not authorized
     */
    async authorizeUpdateAsync(viewerContext, queryContext, entity) {
        return await this.authorizeForRulesetAsync(this.updateRules, viewerContext, queryContext, entity, EntityAuthorizationAction.UPDATE);
    }
    /**
     * Authorize an entity against deletion policy.
     * @param viewerContext - viewer context of user deleting the entity
     * @param queryContext - query context in which to perform the delete authorization
     * @param entity - entity to authorize
     * @returns entity if authorized
     * @throws {@link EntityNotAuthorizedError} when not authorized
     */
    async authorizeDeleteAsync(viewerContext, queryContext, entity) {
        return await this.authorizeForRulesetAsync(this.deleteRules, viewerContext, queryContext, entity, EntityAuthorizationAction.DELETE);
    }
    async authorizeForRulesetAsync(ruleset, viewerContext, queryContext, entity, action) {
        const privacyPolicyEvaluator = this.getPrivacyPolicyEvaluator(viewerContext);
        switch (privacyPolicyEvaluator.mode) {
            case EntityPrivacyPolicyEvaluationMode.ENFORCE:
                return await this.authorizeForRulesetInnerAsync(ruleset, viewerContext, queryContext, entity, action);
            case EntityPrivacyPolicyEvaluationMode.ENFORCE_AND_LOG:
                try {
                    return await this.authorizeForRulesetInnerAsync(ruleset, viewerContext, queryContext, entity, action);
                }
                catch (e) {
                    if (!(e instanceof EntityNotAuthorizedError_1.default)) {
                        throw e;
                    }
                    privacyPolicyEvaluator.denyHandler(e);
                    throw e;
                }
            case EntityPrivacyPolicyEvaluationMode.DRY_RUN:
                try {
                    return await this.authorizeForRulesetInnerAsync(ruleset, viewerContext, queryContext, entity, action);
                }
                catch (e) {
                    if (!(e instanceof EntityNotAuthorizedError_1.default)) {
                        throw e;
                    }
                    privacyPolicyEvaluator.denyHandler(e);
                    return entity;
                }
        }
    }
    async authorizeForRulesetInnerAsync(ruleset, viewerContext, queryContext, entity, action) {
        for (let i = 0; i < ruleset.length; i++) {
            const rule = ruleset[i];
            const ruleEvaluationResult = await rule.evaluateAsync(viewerContext, queryContext, entity);
            switch (ruleEvaluationResult) {
                case PrivacyPolicyRule_1.RuleEvaluationResult.DENY:
                    throw new EntityNotAuthorizedError_1.default(entity, viewerContext, action, i);
                case PrivacyPolicyRule_1.RuleEvaluationResult.SKIP:
                    continue;
                case PrivacyPolicyRule_1.RuleEvaluationResult.ALLOW:
                    return entity;
                default:
                    throw new Error('should not be a fourth type of rule evaluation result');
            }
        }
        throw new EntityNotAuthorizedError_1.default(entity, viewerContext, action, -1);
    }
}
exports.default = EntityPrivacyPolicy;
//# sourceMappingURL=EntityPrivacyPolicy.js.map