import { IEntityClass } from './Entity';
import EntityAssociationLoader from './EntityAssociationLoader';
import EntityLoader from './EntityLoader';
import EntityPrivacyPolicy from './EntityPrivacyPolicy';
import { EntityQueryContext } from './EntityQueryContext';
import ViewerContext from './ViewerContext';
/**
 * A readonly entity exposes only the read functionality of an Entity. Used as the base
 * type for most entity operations excluding mutations.
 *
 * This is also useful as a base class for Entities that should not be mutated, such as:
 * - Entities representing SQL views.
 * - Entities representing immutable tables.
 */
export default abstract class ReadonlyEntity<TFields, TID extends NonNullable<TFields[TSelectedFields]>, TViewerContext extends ViewerContext, TSelectedFields extends keyof TFields = keyof TFields> {
    private readonly viewerContext;
    private readonly databaseFields;
    private readonly id;
    private readonly rawFields;
    /**
     * Constructs an instance of an Entity.
     * @param viewerContext - the ViewerContext reading this entity
     * @param rawFields - all underlying fields for this entity's data
     *
     * @internal
     */
    constructor(viewerContext: TViewerContext, databaseFields: Readonly<TFields>);
    toString(): string;
    getUniqueIdentifier(): string;
    /**
     * @returns the ViewerContext authorized to read this entity
     */
    getViewerContext(): TViewerContext;
    /**
     * @returns the ID of this entity
     */
    getID(): TID;
    /**
     * @returns {@link EntityAssociationLoader} for this entity
     */
    associationLoader(): EntityAssociationLoader<TFields, TID, TViewerContext, this, TSelectedFields>;
    /**
     * Get a underlying field from this entity's data
     * @param fieldName - the field to get
     * @returns the value of the field or undefined if not loaded with that field
     */
    getField<K extends keyof Pick<TFields, TSelectedFields>>(fieldName: K): Pick<TFields, TSelectedFields>[K];
    /**
     * @returns all underlying fields from this entity's data
     */
    getAllFields(): Readonly<Pick<TFields, TSelectedFields>>;
    /**
     * @returns all underlying fields from this entity's database data
     */
    getAllDatabaseFields(): Readonly<TFields>;
    /**
     * Vend loader for loading an entity in a given query context.
     * @param viewerContext - viewer context of loading user
     * @param queryContext - query context in which to perform the load
     */
    static loader<TMFields, TMID extends NonNullable<TMFields[TMSelectedFields]>, TMViewerContext extends ViewerContext, TMViewerContext2 extends TMViewerContext, TMEntity extends ReadonlyEntity<TMFields, TMID, TMViewerContext, TMSelectedFields>, TMPrivacyPolicy extends EntityPrivacyPolicy<TMFields, TMID, TMViewerContext, TMEntity, TMSelectedFields>, TMSelectedFields extends keyof TMFields = keyof TMFields>(this: IEntityClass<TMFields, TMID, TMViewerContext, TMEntity, TMPrivacyPolicy, TMSelectedFields>, viewerContext: TMViewerContext2, queryContext?: EntityQueryContext): EntityLoader<TMFields, TMID, TMViewerContext, TMEntity, TMPrivacyPolicy, TMSelectedFields>;
}
