"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const invariant_1 = __importDefault(require("invariant"));
const EntityAssociationLoader_1 = __importDefault(require("./EntityAssociationLoader"));
const entityUtils_1 = require("./entityUtils");
/**
 * A readonly entity exposes only the read functionality of an Entity. Used as the base
 * type for most entity operations excluding mutations.
 *
 * This is also useful as a base class for Entities that should not be mutated, such as:
 * - Entities representing SQL views.
 * - Entities representing immutable tables.
 */
class ReadonlyEntity {
    /**
     * Constructs an instance of an Entity.
     * @param viewerContext - the ViewerContext reading this entity
     * @param rawFields - all underlying fields for this entity's data
     *
     * @internal
     */
    constructor(viewerContext, databaseFields) {
        this.viewerContext = viewerContext;
        this.databaseFields = databaseFields;
        const companionDefinition = this
            .constructor.getCompanionDefinition();
        const idField = companionDefinition.entityConfiguration.idField;
        const id = databaseFields[idField];
        invariant_1.default(id, 'must provide ID to create an entity');
        this.id = id;
        const entitySelectedFields = companionDefinition.entitySelectedFields;
        this.rawFields = entityUtils_1.pick(databaseFields, entitySelectedFields);
    }
    toString() {
        return `${this.constructor.name}[${this.getID()}]`;
    }
    getUniqueIdentifier() {
        return this.toString();
    }
    /**
     * @returns the ViewerContext authorized to read this entity
     */
    getViewerContext() {
        return this.viewerContext;
    }
    /**
     * @returns the ID of this entity
     */
    getID() {
        return this.id;
    }
    /**
     * @returns {@link EntityAssociationLoader} for this entity
     */
    associationLoader() {
        return new EntityAssociationLoader_1.default(this);
    }
    /**
     * Get a underlying field from this entity's data
     * @param fieldName - the field to get
     * @returns the value of the field or undefined if not loaded with that field
     */
    getField(fieldName) {
        return this.rawFields[fieldName];
    }
    /**
     * @returns all underlying fields from this entity's data
     */
    getAllFields() {
        return { ...this.rawFields };
    }
    /**
     * @returns all underlying fields from this entity's database data
     */
    getAllDatabaseFields() {
        return { ...this.databaseFields };
    }
    /**
     * Vend loader for loading an entity in a given query context.
     * @param viewerContext - viewer context of loading user
     * @param queryContext - query context in which to perform the load
     */
    static loader(viewerContext, queryContext = viewerContext
        .getViewerScopedEntityCompanionForClass(this)
        .getQueryContextProvider()
        .getQueryContext()) {
        return viewerContext
            .getViewerScopedEntityCompanionForClass(this)
            .getLoaderFactory()
            .forLoad(queryContext);
    }
}
exports.default = ReadonlyEntity;
//# sourceMappingURL=ReadonlyEntity.js.map