"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const Entity_1 = __importDefault(require("../Entity"));
const EntityCompanionProvider_1 = require("../EntityCompanionProvider");
const EntityConfiguration_1 = __importDefault(require("../EntityConfiguration"));
const EntityFields_1 = require("../EntityFields");
const EntityPrivacyPolicy_1 = __importDefault(require("../EntityPrivacyPolicy"));
const ReadThroughEntityCache_1 = require("../internal/ReadThroughEntityCache");
const AlwaysAllowPrivacyPolicyRule_1 = __importDefault(require("../rules/AlwaysAllowPrivacyPolicyRule"));
const TestViewerContext_1 = __importDefault(require("../testfixtures/TestViewerContext"));
const createUnitTestEntityCompanionProvider_1 = require("../utils/testing/createUnitTestEntityCompanionProvider");
class TestEntityPrivacyPolicy extends EntityPrivacyPolicy_1.default {
    constructor() {
        super(...arguments);
        this.readRules = [
            new AlwaysAllowPrivacyPolicyRule_1.default(),
        ];
        this.createRules = [
            new AlwaysAllowPrivacyPolicyRule_1.default(),
        ];
        this.updateRules = [
            new AlwaysAllowPrivacyPolicyRule_1.default(),
        ];
        this.deleteRules = [
            new AlwaysAllowPrivacyPolicyRule_1.default(),
        ];
    }
}
// eslint-disable-next-line @typescript-eslint/explicit-function-return-type
const makeEntityClasses = (edgeDeletionBehavior) => {
    class ParentEntity extends Entity_1.default {
        static getCompanionDefinition() {
            return parentEntityCompanion;
        }
    }
    class ChildEntity extends Entity_1.default {
        static getCompanionDefinition() {
            return childEntityCompanion;
        }
    }
    class GrandChildEntity extends Entity_1.default {
        static getCompanionDefinition() {
            return grandChildEntityCompanion;
        }
    }
    const parentEntityConfiguration = new EntityConfiguration_1.default({
        idField: 'id',
        tableName: 'parents',
        getInboundEdges: () => [ChildEntity],
        schema: {
            id: new EntityFields_1.UUIDField({
                columnName: 'id',
                cache: true,
            }),
        },
        databaseAdapterFlavor: 'postgres',
        cacheAdapterFlavor: 'redis',
    });
    const childEntityConfiguration = new EntityConfiguration_1.default({
        idField: 'id',
        tableName: 'children',
        getInboundEdges: () => [GrandChildEntity],
        schema: {
            id: new EntityFields_1.UUIDField({
                columnName: 'id',
                cache: true,
            }),
            parent_id: new EntityFields_1.UUIDField({
                columnName: 'parent_id',
                cache: true,
                association: {
                    getAssociatedEntityClass: () => ParentEntity,
                    edgeDeletionBehavior,
                },
            }),
        },
        databaseAdapterFlavor: 'postgres',
        cacheAdapterFlavor: 'redis',
    });
    const grandChildEntityConfiguration = new EntityConfiguration_1.default({
        idField: 'id',
        tableName: 'grandchildren',
        schema: {
            id: new EntityFields_1.UUIDField({
                columnName: 'id',
                cache: true,
            }),
            parent_id: new EntityFields_1.UUIDField({
                columnName: 'parent_id',
                cache: true,
                association: {
                    getAssociatedEntityClass: () => ChildEntity,
                    edgeDeletionBehavior,
                },
            }),
        },
        databaseAdapterFlavor: 'postgres',
        cacheAdapterFlavor: 'redis',
    });
    const parentEntityCompanion = new EntityCompanionProvider_1.EntityCompanionDefinition({
        entityClass: ParentEntity,
        entityConfiguration: parentEntityConfiguration,
        privacyPolicyClass: TestEntityPrivacyPolicy,
    });
    const childEntityCompanion = new EntityCompanionProvider_1.EntityCompanionDefinition({
        entityClass: ChildEntity,
        entityConfiguration: childEntityConfiguration,
        privacyPolicyClass: TestEntityPrivacyPolicy,
    });
    const grandChildEntityCompanion = new EntityCompanionProvider_1.EntityCompanionDefinition({
        entityClass: GrandChildEntity,
        entityConfiguration: grandChildEntityConfiguration,
        privacyPolicyClass: TestEntityPrivacyPolicy,
    });
    return {
        ParentEntity,
        ChildEntity,
        GrandChildEntity,
    };
};
describe('EntityMutator.processEntityDeletionForInboundEdgesAsync', () => {
    describe('EntityEdgeDeletionBehavior.CASCADE_DELETE', () => {
        it('deletes', async () => {
            const { ParentEntity, ChildEntity, GrandChildEntity } = makeEntityClasses(EntityFields_1.EntityEdgeDeletionBehavior.CASCADE_DELETE);
            const companionProvider = createUnitTestEntityCompanionProvider_1.createUnitTestEntityCompanionProvider();
            const viewerContext = new TestViewerContext_1.default(companionProvider);
            const parent = await ParentEntity.creator(viewerContext).enforceCreateAsync();
            const child = await ChildEntity.creator(viewerContext)
                .setField('parent_id', parent.getID())
                .enforceCreateAsync();
            const grandchild = await GrandChildEntity.creator(viewerContext)
                .setField('parent_id', child.getID())
                .enforceCreateAsync();
            await expect(ParentEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(parent.getID())).resolves.not.toBeNull();
            await expect(ChildEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(child.getID())).resolves.not.toBeNull();
            await expect(GrandChildEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(grandchild.getID())).resolves.not.toBeNull();
            await ParentEntity.enforceDeleteAsync(parent);
            await expect(ParentEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(parent.getID())).resolves.toBeNull();
            await expect(ChildEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(child.getID())).resolves.toBeNull();
            await expect(GrandChildEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(grandchild.getID())).resolves.toBeNull();
        });
    });
    describe('EntityEdgeDeletionBehavior.SET_NULL', () => {
        it('sets null', async () => {
            const { ParentEntity, ChildEntity, GrandChildEntity } = makeEntityClasses(EntityFields_1.EntityEdgeDeletionBehavior.SET_NULL);
            const companionProvider = createUnitTestEntityCompanionProvider_1.createUnitTestEntityCompanionProvider();
            const viewerContext = new TestViewerContext_1.default(companionProvider);
            const parent = await ParentEntity.creator(viewerContext).enforceCreateAsync();
            const child = await ChildEntity.creator(viewerContext)
                .setField('parent_id', parent.getID())
                .enforceCreateAsync();
            const grandchild = await GrandChildEntity.creator(viewerContext)
                .setField('parent_id', child.getID())
                .enforceCreateAsync();
            await expect(ParentEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(parent.getID())).resolves.not.toBeNull();
            await expect(ChildEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(child.getID())).resolves.not.toBeNull();
            await expect(GrandChildEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(grandchild.getID())).resolves.not.toBeNull();
            await ParentEntity.enforceDeleteAsync(parent);
            await expect(ParentEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(parent.getID())).resolves.toBeNull();
            const loadedChild = await ChildEntity.loader(viewerContext)
                .enforcing()
                .loadByIDAsync(child.getID());
            expect(loadedChild.getField('parent_id')).toBeNull();
            const loadedGrandchild = await GrandChildEntity.loader(viewerContext)
                .enforcing()
                .loadByIDAsync(grandchild.getID());
            expect(loadedGrandchild.getField('parent_id')).toEqual(loadedChild.getID());
        });
    });
    describe('EntityEdgeDeletionBehavior.CASCADE_DELETE_INVALIDATE_CACHE', () => {
        it('invalidates the cache', async () => {
            var _a, _b, _c, _d;
            const { ParentEntity, ChildEntity, GrandChildEntity } = makeEntityClasses(EntityFields_1.EntityEdgeDeletionBehavior.CASCADE_DELETE_INVALIDATE_CACHE);
            const companionProvider = createUnitTestEntityCompanionProvider_1.createUnitTestEntityCompanionProvider();
            const viewerContext = new TestViewerContext_1.default(companionProvider);
            const parent = await ParentEntity.creator(viewerContext).enforceCreateAsync();
            const child = await ChildEntity.creator(viewerContext)
                .setField('parent_id', parent.getID())
                .enforceCreateAsync();
            const grandchild = await GrandChildEntity.creator(viewerContext)
                .setField('parent_id', child.getID())
                .enforceCreateAsync();
            await expect(ParentEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(parent.getID())).resolves.not.toBeNull();
            await expect(ChildEntity.loader(viewerContext)
                .enforcing()
                .loadByFieldEqualingAsync('parent_id', parent.getID())).resolves.not.toBeNull();
            await expect(GrandChildEntity.loader(viewerContext)
                .enforcing()
                .loadByFieldEqualingAsync('parent_id', child.getID())).resolves.not.toBeNull();
            const childCacheAdapter = viewerContext.getViewerScopedEntityCompanionForClass(ChildEntity)['entityCompanion']['tableDataCoordinator']['cacheAdapter'];
            const childCachedBefore = await childCacheAdapter.loadManyAsync('parent_id', [
                parent.getID(),
            ]);
            expect((_a = childCachedBefore.get(parent.getID())) === null || _a === void 0 ? void 0 : _a.status).toEqual(ReadThroughEntityCache_1.CacheStatus.HIT);
            const grandChildCacheAdapter = viewerContext.getViewerScopedEntityCompanionForClass(GrandChildEntity)['entityCompanion']['tableDataCoordinator']['cacheAdapter'];
            const grandChildCachedBefore = await grandChildCacheAdapter.loadManyAsync('parent_id', [
                child.getID(),
            ]);
            expect((_b = grandChildCachedBefore.get(child.getID())) === null || _b === void 0 ? void 0 : _b.status).toEqual(ReadThroughEntityCache_1.CacheStatus.HIT);
            await ParentEntity.enforceDeleteAsync(parent);
            const childCachedAfter = await childCacheAdapter.loadManyAsync('parent_id', [parent.getID()]);
            expect((_c = childCachedAfter.get(parent.getID())) === null || _c === void 0 ? void 0 : _c.status).toEqual(ReadThroughEntityCache_1.CacheStatus.MISS);
            const grandChildCachedAfter = await grandChildCacheAdapter.loadManyAsync('parent_id', [
                child.getID(),
            ]);
            expect((_d = grandChildCachedAfter.get(child.getID())) === null || _d === void 0 ? void 0 : _d.status).toEqual(ReadThroughEntityCache_1.CacheStatus.MISS);
            await expect(ParentEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(parent.getID())).resolves.toBeNull();
            await expect(ChildEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(child.getID())).resolves.not.toBeNull();
            await expect(GrandChildEntity.loader(viewerContext).enforcing().loadByIDNullableAsync(grandchild.getID())).resolves.not.toBeNull();
        });
    });
});
//# sourceMappingURL=EntityEdges-test.js.map